#include "GOLFThreeHarveyBackgroundModel.h"


// GOLFThreeHarveyBackgroundModel::GOLFThreeHarveyBackgroundModel()
//
// PURPOSE: 
//      Constructor. Initializes model computation.
//
// INPUT:
//      covariates:             one-dimensional array containing the values
//                              of the independent variable.
//

GOLFThreeHarveyBackgroundModel::GOLFThreeHarveyBackgroundModel(const RefArrayXd covariates)
: BackgroundModel(covariates)
{
    // Create response function modulating the sampling rate of input Kepler LC data

    // double NyquistFrequency = 8333.3338;    // muHz         GOLF at 60 s sampling
    // double NyquistFrequency = 25000.000;    // muHz         GOLF at 20 s sampling
    double NyquistFrequency = 12500.000;    // muHz         GOLF at 40 s sampling
    ArrayXd sincFunctionArgument = (Functions::PI / 2.0) * covariates / NyquistFrequency;
    responseFunction = (sincFunctionArgument.sin() / sincFunctionArgument).square(); 
}










// GOLFThreeHarveyBackgroundModel::GOLFThreeHarveyBackgroundModel()
//
// PURPOSE: 
//      Destructor.
//

GOLFThreeHarveyBackgroundModel::~GOLFThreeHarveyBackgroundModel()
{

}










// GOLFThreeHarveyBackgroundModel::predict()
//
// PURPOSE:
//      Builds the predictions from a background model for GOLF data.
//      The model consists of one constant component, three Harvey-like profiles
//      and two Lorentzians for modeling the oscillation envelope.
//
// INPUT:
//      predictions:        one-dimensional array to contain the predictions
//                          from the model
//      configuringParameters:    one-dimensional array where each element
//                          contains the value of a free parameter of the model
//
// OUTPUT:
//      void
//
// NOTE:
//      The free parameters are to be given in the order
//      (1) White noise background (flat noise level, (m/s)^2 / muHz)
//      (2) Amplitude of the first background component (m/s)
//      (3) Frequency of the first background component (muHz)
//      (4) Exponent of the first background component
//      (5) Amplitude of the second background component (m/s)
//      (6) Frequency of the second background component (muHz)
//      (7) Exponent of the second background component
//      (8) Amplitude of the third background component (m/s)
//      (9) Frequency of the third background component (muHz)
//      (10) Exponent of the third background component
//      (11) Height first Lorentzian ((m/s)^2 / muHz)
//      (12) Frequency first Lorentzian (muHz)
//      (13) Width first Lorentzian (muHz)
//      (14) Exponent first Lorentzian
//      (15) Height second Lorentzian ((m/s)^2 / muHz)
//      (16) Frequency second Lorentzian (muHz)
//      (17) Width second Lorentzian (muHz)
//      (18) Exponent second Lorentzian
//

void GOLFThreeHarveyBackgroundModel::predict(RefArrayXd predictions)
{
    Nparameters = configuringParameters.size();


    // Initialize global parameters

    double flatNoiseLevel = configuringParameters(0);
    double amplitudeHarvey1 = configuringParameters(1);
    double frequencyHarvey1 = configuringParameters(2);
    double exponentHarvey1 = configuringParameters(3);
    double amplitudeHarvey2 = configuringParameters(4);
    double frequencyHarvey2 = configuringParameters(5);
    double exponentHarvey2 = configuringParameters(6);  
    double amplitudeHarvey3 = configuringParameters(7);
    double frequencyHarvey3 = configuringParameters(8);
    double exponentHarvey3 = configuringParameters(9); 


    // Compute Harvey-like components and add them to the predictions
    
    double beta = 2.0/Functions::PI;
    predictions += beta*amplitudeHarvey1*amplitudeHarvey1/(frequencyHarvey1*(1.0 + (covariates/frequencyHarvey1).pow(exponentHarvey1)));
    predictions += beta*amplitudeHarvey2*amplitudeHarvey2/(frequencyHarvey2*(1.0 + (covariates/frequencyHarvey2).pow(exponentHarvey2)));
    predictions += beta*amplitudeHarvey3*amplitudeHarvey3/(frequencyHarvey3*(1.0 + (covariates/frequencyHarvey3).pow(exponentHarvey3)));


    // Modulate the model by the response function (apodization)
    
    predictions *= responseFunction;           


    // Add flat noise level component

    predictions += flatNoiseLevel;
}











