// Main code for peak bagging of the solar data VIRGO SPM by means of nested sampling analysis
// Created by Enrico Corsaro @ CEA - November 2015
// e-mail: emncorsaro@gmail.com
// Source code file "SunGOLFPeakBagging.cpp"

#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <Eigen/Dense>
#include "Functions.h"
#include "File.h"
#include "MultiEllipsoidSampler.h"
#include "KmeansClusterer.h"
#include "EuclideanMetric.h"
#include "Prior.h"
#include "UniformPrior.h"
#include "NormalPrior.h"
#include "ExponentialLikelihood.h"
#include "LorentzianRotationMixtureModel.h"
#include "VIRGOThreeHarveyBackgroundModel.h"
#include "PowerlawReducer.h"
#include "Results.h"
#include "Ellipsoid.h"

int main(int argc, char *argv[])
{
    // Check number of arguments for main function
    
    if (argc != 3)
    {
        cerr << "Usage: sunpeakbagging <year number> <run number>" << endl;
        exit(EXIT_FAILURE);
    }


    // ---------------------------
    // ----- Read input data -----
    // ---------------------------

    unsigned long Nrows;
    int Ncols;
    ArrayXXd data;

    string yearNumber(argv[1]);
    string runNumber(argv[2]);
    string baseInputDirName = "***YOUR LOCAL PAATH HERE***/SunPeakBagging/data/VIRGO/";
    string inputFileName = baseInputDirName + "Year" + yearNumber + ".txt";
        
    string outputDirName = "***YOUR LOCAL PAATH HERE***/SunPeakBagging/results/VIRGO/";
    outputDirName += "Year" + yearNumber + "/ThreeHarvey/";
    string outputPathPrefix = outputDirName + runNumber + "/peakbagging_";

    cerr << "Input filename: " << inputFileName << endl;
    cerr << "Output path prefix: " << outputPathPrefix << endl;

    ifstream inputFile;
    File::openInputFile(inputFile, inputFileName);
    File::sniffFile(inputFile, Nrows, Ncols);
    data = File::arrayXXdFromFile(inputFile, Nrows, Ncols);
    inputFile.close();

   
    // Creating arrays for each data type
    
    ArrayXd covariates = data.col(0);
    ArrayXd observations = data.col(1); 
   

    // Read input frequency range of the PSD
    
    inputFileName = outputDirName + "frequencyRange.txt";
    File::openInputFile(inputFile, inputFileName);
    File::sniffFile(inputFile, Nrows, Ncols);
    ArrayXd frequencyRange;
    frequencyRange = File::arrayXXdFromFile(inputFile, Nrows, Ncols);
    inputFile.close();

    double lowerFrequency = frequencyRange(0);        // muHz
    double upperFrequency = frequencyRange(1);        // muHz


    // Trim input dataset in the given frequency range
    
    vector<int> trimIndices = Functions::findArrayIndicesWithinBoundaries(covariates, lowerFrequency, upperFrequency);
    int Nbins = trimIndices.size();
    ArrayXd trimmedArray(Nbins);
    trimmedArray = covariates.segment(trimIndices[0],Nbins);
    covariates.resize(Nbins);
    covariates = trimmedArray;
    trimmedArray = observations.segment(trimIndices[0],Nbins);
    observations.resize(Nbins);
    observations = trimmedArray;

    cout << endl;
    cout << "Frequency range: [" 
         << setprecision(4) 
         << covariates.minCoeff() 
         << ", " 
         << covariates.maxCoeff() 
         << "] muHz" << endl;
    cout << endl;

    
    // -------------------------------------------------------
    // ----- First step. Set up all prior distributions -----
    // -------------------------------------------------------
    
    unsigned long Nparameters;              // Number of parameters for which prior distributions are defined
    
    // ---- Read prior hyper parameters for resolved modes -----
    
    inputFileName = outputDirName + "modesWithRotation_hyperParameters.txt";
    File::openInputFile(inputFile, inputFileName);
    File::sniffFile(inputFile, Nparameters, Ncols);
    ArrayXXd hyperParameters;
    
    if (fmod(Nparameters,4) != 0.0)
    {
        cerr << "Wrong number of input prior hyper-parameters for resolved peaks." << endl;
        exit(EXIT_FAILURE);
    }
   
    if (Ncols == 1)
    {
        Ncols = 2;
        hyperParameters.conservativeResize(Nparameters,Ncols);
    }

    hyperParameters = File::arrayXXdFromFile(inputFile, Nparameters, Ncols);
    inputFile.close();
    
    int NmodesWithRotation = Nparameters/4;
    ArrayXd hyperParametersMinima1 = hyperParameters.col(0);
    ArrayXd hyperParametersMaxima1 = hyperParameters.col(1);
    // ---------------------------------------------------------


    // --- Read prior hyper parameters for unresolved modes ----
    
    inputFileName = outputDirName + "modesWithoutRotation_hyperParameters.txt";
    File::openInputFile(inputFile, inputFileName);
    File::sniffFile(inputFile, Nparameters, Ncols);
  
    if (fmod(Nparameters,3) != 0.0)
    {
        cerr << "Wrong number of input prior hyper-parameters for unresolved peaks." << endl;
        exit(EXIT_FAILURE);
    }

    if (Ncols == 1)
    {
        Ncols = 2;
        hyperParameters.conservativeResize(Nparameters,Ncols);
    }
   
    hyperParameters.setZero();
    hyperParameters = File::arrayXXdFromFile(inputFile, Nparameters, Ncols);
    inputFile.close();

    int NmodesWithoutRotation = Nparameters/3;
    ArrayXd hyperParametersMinima2 = hyperParameters.col(0);
    ArrayXd hyperParametersMaxima2 = hyperParameters.col(1);
    // ---------------------------------------------------------
   

    int Ndimensions = 4*NmodesWithRotation + 3*NmodesWithoutRotation;   // Total number of dimensions of the peak bagging model


    // Uniform Prior
    
    int NpriorTypes = 1;                                        // Total number of prior types included in the computation
    vector<Prior*> ptrPriors(NpriorTypes);
    
    ArrayXd parametersMinima(Ndimensions);                      // Minima
    ArrayXd parametersMaxima(Ndimensions);                      // Maxima
    parametersMinima << hyperParametersMinima1, hyperParametersMinima2;
    parametersMaxima << hyperParametersMaxima1, hyperParametersMaxima2; 
    UniformPrior uniformPrior(parametersMinima, parametersMaxima);
    ptrPriors[0] = &uniformPrior;
    

    // -------------------------------------------------------------------
    // ---- Second step. Set up the models for the inference problem ----- 
    // -------------------------------------------------------------------
    
    // Chose a model for background and configure it, then do the same for the peakbagging model.
    
    VIRGOThreeHarveyBackgroundModel backgroundModel(covariates);             // 1 Lorentzian + 3 Harvey-like components
    string backgroundConfiguringParameters = outputDirName + "backgroundParameters.txt";
    backgroundModel.readConfiguringParametersFromFile(backgroundConfiguringParameters);
    
    double cosi = 0.0;      // Solar inclination angle of observation i = 90 degrees
    string angularDegreesFileName = outputDirName + "modesWithRotation_angularDegrees.txt";     // THe angular degrees of the rotational split modes
    LorentzianRotationMixtureModel model(covariates, NmodesWithRotation, NmodesWithoutRotation, cosi, angularDegreesFileName, backgroundModel);


    // -----------------------------------------------------------------
    // ----- Third step. Set up the likelihood function to be used -----
    // -----------------------------------------------------------------
    
    ExponentialLikelihood likelihood(observations, model);
    

    // -------------------------------------------------------------------------------
    // ----- Fourth step. Set up the X-means clusterer using an Euclidean metric -----
    // -------------------------------------------------------------------------------

    inputFileName = outputDirName + "Xmeans_configuringParameters.txt";
    File::openInputFile(inputFile, inputFileName);
    File::sniffFile(inputFile, Nparameters, Ncols);

    if (Nparameters != 2)
    {
        cerr << "Wrong number of input parameters for X-means algorithm." << endl;
        exit(EXIT_FAILURE);
    }

    ArrayXd configuringParameters;
    configuringParameters = File::arrayXXdFromFile(inputFile, Nparameters, Ncols);
    inputFile.close();
    
    int minNclusters = configuringParameters(0);
    int maxNclusters = configuringParameters(1);
    
    if ((minNclusters <= 0) || (maxNclusters <= 0) || (maxNclusters < minNclusters))
    {
        cerr << "Minimum or maximum number of clusters cannot be <= 0, and " << endl;
        cerr << "minimum number of clusters cannot be larger than maximum number of clusters." << endl;
        exit(EXIT_FAILURE);
    }
    
    int Ntrials = 10;
    double relTolerance = 0.01;
   
    EuclideanMetric myMetric;
    KmeansClusterer kmeans(myMetric, minNclusters, maxNclusters, Ntrials, relTolerance); 


    // ---------------------------------------------------------------------
    // ----- Sixth step. Configure and start nested sampling inference -----
    // ---------------------------------------------------------------------

    inputFileName = outputDirName + "NSMC_configuringParameters.txt";
    File::openInputFile(inputFile, inputFileName);
    File::sniffFile(inputFile, Nparameters, Ncols);
    configuringParameters.setZero();
    configuringParameters = File::arrayXXdFromFile(inputFile, Nparameters, Ncols);
    inputFile.close();

    if (Nparameters != 7)
    {
        cerr << "Wrong number of input parameters for NSMC algorithm." << endl;
        exit(EXIT_FAILURE);
    }
    

    bool printOnTheScreen = true;                       // Print results on the screen
    int initialNobjects = configuringParameters(0);     // Initial number of live points 
    int minNobjects = configuringParameters(1);         // Minimum number of live points 
    int maxNdrawAttempts = configuringParameters(2);    // Maximum number of attempts when trying to draw a new sampling point
    int NinitialIterationsWithoutClustering = configuringParameters(3); // The first N iterations, we assume that there is only 1 cluster
    int NiterationsWithSameClustering = configuringParameters(4);       // Clustering is only happening every N iterations.
    double initialEnlargementFraction = 0.267*pow(Ndimensions,0.643);   // Fraction by which each axis in an ellipsoid has to be enlarged.
                                                                        // It can be a number >= 0, where 0 means no enlargement.
    double shrinkingRate = configuringParameters(5);        // Exponent for remaining prior mass in ellipsoid enlargement fraction.
                                                            // It is a number between 0 and 1. The smaller the slower the shrinkage
                                                            // of the ellipsoids.
                                                            
    if ((shrinkingRate > 1) || (shrinkingRate) < 0)
    {
        cerr << "Shrinking Rate for ellipsoids must be in the range [0, 1]. " << endl;
        exit(EXIT_FAILURE);
    }

    double terminationFactor = configuringParameters(6);    // Termination factor for nested sampling process.

    MultiEllipsoidSampler nestedSampler(printOnTheScreen, ptrPriors, likelihood, myMetric, kmeans, 
                                        initialNobjects, minNobjects, initialEnlargementFraction, shrinkingRate);
    
    double tolerance = 1.e2;
    double exponent = 0.4;
    PowerlawReducer livePointsReducer(nestedSampler, tolerance, exponent, terminationFactor);

    nestedSampler.run(livePointsReducer, NinitialIterationsWithoutClustering, NiterationsWithSameClustering, 
                      maxNdrawAttempts, terminationFactor, outputPathPrefix);

    nestedSampler.outputFile << "# List of configuring parameters used for the ellipsoidal sampler and X-means" << endl;
    nestedSampler.outputFile << "# Row #1: Minimum Nclusters" << endl;
    nestedSampler.outputFile << "# Row #2: Maximum Nclusters" << endl;
    nestedSampler.outputFile << "# Row #3: Initial Enlargement Fraction" << endl;
    nestedSampler.outputFile << "# Row #4: Shrinking Rate" << endl;
    nestedSampler.outputFile << minNclusters << endl;
    nestedSampler.outputFile << maxNclusters << endl;
    nestedSampler.outputFile << initialEnlargementFraction << endl;
    nestedSampler.outputFile << shrinkingRate << endl;
    nestedSampler.outputFile.close();


    // -------------------------------------------------------
    // ----- Last step. Save the results in output files -----
    // -------------------------------------------------------
   
    Results results(nestedSampler);
    results.writeParametersToFile("parameter");
    results.writeLogLikelihoodToFile("logLikelihood.txt");
    results.writeEvidenceInformationToFile("evidenceInformation.txt");
    results.writePosteriorProbabilityToFile("posteriorDistribution.txt");

    double credibleLevel = 68.3;
    bool writeMarginalDistributionToFile = true;
    results.writeParametersSummaryToFile("parameterSummary.txt", credibleLevel, writeMarginalDistributionToFile);
 
    cerr << "Process #" << runNumber << " has been completed." << endl;

    return EXIT_SUCCESS;
}
